package at.ac.uibk.gitsearch.repository.search.testESService;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringWriter;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLEncoder;
import java.nio.charset.Charset;
import java.nio.file.Path;
import java.time.Instant;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.stream.Collectors;

import org.apache.commons.lang3.StringUtils;
import org.apache.http.HttpHost;
import org.apache.http.util.EntityUtils;
import org.apache.logging.log4j.LogManager;
import org.codeability.sharing.plugins.api.search.SearchResultDTO;
import org.codeability.sharing.plugins.api.search.SearchResultDTO.GitProject;
import org.codeability.sharing.plugins.api.search.SearchResultDTO.MetadataFile;
import org.codeability.sharing.plugins.api.search.UserProvidedMetadataDTO;
import org.codeability.sharing.plugins.api.search.util.ExerciseId;
import org.elasticsearch.action.search.SearchRequest;
import org.elasticsearch.action.search.SearchResponse;
import org.elasticsearch.client.RequestOptions;
import org.elasticsearch.client.Response;
import org.elasticsearch.client.RestClient;
import org.elasticsearch.client.RestHighLevelClient;
import org.elasticsearch.common.settings.Settings;
import org.elasticsearch.index.query.BoolQueryBuilder;
import org.elasticsearch.index.query.QueryBuilders;
import org.elasticsearch.node.InternalSettingsPreparer;
import org.elasticsearch.node.Node;
import org.elasticsearch.node.NodeValidationException;
import org.elasticsearch.plugins.Plugin;
import org.elasticsearch.search.builder.SearchSourceBuilder;
import org.elasticsearch.transport.Netty4Plugin;
import org.slf4j.LoggerFactory;
import org.springframework.boot.context.event.ApplicationStartingEvent;
import org.springframework.context.event.ContextStartedEvent;
import org.springframework.context.event.ContextStoppedEvent;
import org.springframework.context.event.EventListener;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.MediaType;
import org.springframework.util.Assert;
import org.springframework.web.client.RestClientException;
import org.springframework.web.client.RestTemplate;

import com.fasterxml.jackson.core.JsonGenerationException;
import com.fasterxml.jackson.core.JsonParseException;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.dataformat.yaml.YAMLFactory;
import com.fasterxml.jackson.datatype.jsr310.JavaTimeModule;
import com.google.common.base.Charsets;
import com.google.common.io.ByteSource;

import at.ac.uibk.gitsearch.repository.search.MetaDataRepository;
import at.ac.uibk.gitsearch.repository.search.SearchRepositoryConstants;

/**
 * this class provides test infrastructure for a local elastic search service
 * for local tests.
 * 
 * @author Michael Breu
 *
 */
public class ElasticSearchTestServerConfiguration {

	/**
	 * just a helper class to store also data that is not exported to the user
	 * 
	 * @author Michael Breu
	 *
	 */
	public static class ExtendedGitProject extends GitProject {
		private String visibility;
		private boolean archived;
		private int star_count;
		private int open_issues_count;
		private int forks_count;
		private String description;

		/**
		 * @return the visibility
		 */
		public String getVisibility() {
			return visibility;
		}

		/**
		 * @param visibility the visibility to set
		 */
		public void setVisibility(String visibility) {
			this.visibility = visibility;
		}

		/**
		 * @return the archived
		 */
		public boolean isArchived() {
			return archived;
		}

		/**
		 * @param archived the archived to set
		 */
		public void setArchived(boolean archived) {
			this.archived = archived;
		}

		/**
		 * @return the star_count
		 */
		public int getStar_count() {
			return star_count;
		}

		/**
		 * @param star_count the star_count to set
		 */
		public void setStar_count(int star_count) {
			this.star_count = star_count;
		}

		/**
		 * @return the open_issues_count
		 */
		public int getOpen_issues_count() {
			return open_issues_count;
		}

		/**
		 * @param open_issues_count the open_issues_count to set
		 */
		public void setOpen_issues_count(int open_issues_count) {
			this.open_issues_count = open_issues_count;
		}

		/**
		 * @return the forks_count
		 */
		public int getForks_count() {
			return forks_count;
		}

		/**
		 * @param forks_count the forks_count to set
		 */
		public void setForks_count(int forks_count) {
			this.forks_count = forks_count;
		}

		/**
		 * @return the description
		 */
		public String getDescription() {
			return description;
		}

		/**
		 * @param description the description to set
		 */
		public void setDescription(String description) {
			this.description = description;
		}

	}

	private static final org.slf4j.Logger LOGGER = LoggerFactory.getLogger(ElasticSearchTestServerConfiguration.class);
	private TestESNode testNode;

	private ElasticSearchTestServerConfiguration() {
	}

	private static ElasticSearchTestServerConfiguration instance = null;

	public static ElasticSearchTestServerConfiguration getTestInstance() {
		if (instance == null)
			instance = new ElasticSearchTestServerConfiguration();
		return instance;

	}

	public synchronized void startTestNode() throws IOException, NodeValidationException {
		if (testNode == null) {
			LOGGER.info("Starting ES Test Node");
			try {
				File f = File.createTempFile("tobedeleted", "txt");
				Settings s = Settings.builder().put("http.port", "29200").put("transport.port", "29300")
						.put("node.name", "esSharingTest").put("transport.type", "netty4").put("http.type", "netty4")
						.put("path.home", f.getParent() + File.separator + "esTest").build();
				f.delete();

				testNode = new TestESNode(s, Collections.singleton(Netty4Plugin.class)) {

				};
				testNode.start();

				setUpMetaDataIndex();
				setUpContent();
				LOGGER.info("Started ES Test Node");
			} catch (Throwable t) {
				LOGGER.info("ES Test Node crashed: ", t);
			}
		} else {
			LOGGER.info("Node already started");
		}
	}

	private void setUpMetaDataIndex() throws IOException {

		RestTemplate restTemplate = new RestTemplate();
		HttpHeaders headers = new HttpHeaders();
		headers.setContentType(MediaType.APPLICATION_JSON);

		try {
			restTemplate.delete("http://localhost:29200/_all");
		} catch (Throwable e) {
			LOGGER.info("Deletion of previous index failed: {}!", e.getMessage());
		}

		List<String> indices2 = getIndices();

		indices2.forEach((s) -> {
			LOGGER.info("Found index {}", s);
			try {
				restTemplate.delete("http://localhost:29200/" + s, String.class); // just in case, delete old index
			} catch (RestClientException e) {
				LOGGER.info("Deletion of previous index failed: {}!", e.getMessage());
			}
		});

//	    GetIndexRequest request2 = new GetIndexRequest();
//	 
//	    GetIndexResponse response = client.indices().get(request2, RequestOptions.DEFAULT);
//	    String[] indices = response.getIndices();

		HttpEntity<String> request = new HttpEntity<String>(getMetaDataConfigDefinition(), headers);

		restTemplate.put("http://localhost:29200/metadata", request, String.class);

	}

	@SuppressWarnings("deprecation")
	private List<String> getIndices() throws IOException, JsonProcessingException, JsonMappingException {
		List<String> indices2 = null;
		RestHighLevelClient client = new RestHighLevelClient(RestClient.builder(new HttpHost("localhost", 29200)));

		HttpHeaders headers = new HttpHeaders();
		headers.setContentType(MediaType.APPLICATION_JSON);

		RestClient restClient = client.getLowLevelClient();
		Response response2 = null;
		try {
			response2 = restClient.performRequest("GET", "/_cat/indices?v&format=json");
		} catch (IOException e) {
			LOGGER.warn(e.toString(), e);
		}

		// parse the JSON response
		List<HashMap<String, String>> list = null;
		if (response2 != null) {
			ObjectMapper mapper = new ObjectMapper();
			String rawBody = EntityUtils.toString(response2.getEntity());
			TypeReference<List<HashMap<String, String>>> typeRef = new TypeReference<List<HashMap<String, String>>>() {
			};
			list = mapper.readValue(rawBody, typeRef);
		}

		// get the index names
		if (list != null) {
			indices2 = list.stream().map(x -> x.get("index")).collect(Collectors.toList());
		}
		return indices2;
	}

	int contentCount = 0;

	private void setUpContent() throws IOException, URISyntaxException {
		URL test = this.getClass().getResource("../../../service/testData");
		File rootData = new File(test.toURI());
		File metaDataTestDir = new File(rootData, "metaData");

		contentCount = 0;
		visitSubdirectories(metaDataTestDir, (File metaDataFile, List<SearchResultDTO> children, String relativePath,
				ExtendedGitProject gitProject) -> {
			try {
				final UserProvidedMetadataDTO userMetaData = parseMetaDataFile(metaDataFile);

				SearchResultDTO toIndex = new SearchResultDTO();
				toIndex.setMetadata(userMetaData);

				final ExerciseId exerciseId = calculateProjectId(relativePath);

				if (gitProject == null) {
					gitProject = new ExtendedGitProject();

					gitProject.setProjectId(Integer.parseInt(exerciseId.getProjectId()));
					gitProject.setProject_name("SomeGitProject" + exerciseId.getProjectId());
					gitProject.setVisibility("public");

				}

				toIndex.setProject(gitProject);

				toIndex.setExerciseId(exerciseId.toString());

				final MetadataFile file = new MetadataFile();
				file.setFilename(metaDataFile.getName());
				// path currently starts with project id
				int slashPos = relativePath.indexOf("/");
				String projectRelativePath = (slashPos < 0)?metaDataFile.getName():(relativePath.substring(slashPos+1)+"/"+metaDataFile.getName());
				file.setPath(projectRelativePath);
				file.setIndexing_date(Instant.now());
				file.setCommit_id("Unused Commit Id");
				file.setChildren(children.stream().map(c -> c.getExerciseId()).collect(Collectors.toList())
						.toArray(new String[] {}));
				toIndex.setFile(file);

				toIndex.setExerciseId(exerciseId.toString());
				addContentToIndex(toIndex, exerciseId);
				ElasticSearchTestServerConfiguration.this.contentCount++;
				return Collections.singletonList(toIndex);
			} catch (IOException e) {
				e.printStackTrace();
				return null;
			}
		}, "", null /* no initial project */
		);
//	    
//	    
//	    
//	    while(true) {
//	    	contentCount++;
//		    final String content = getContent(contentCount);
//		    if(content==null) break;
//	    }

		waitForCleanStartUp(contentCount - 1);
	}

	private void addContentToIndex(SearchResultDTO content, ExerciseId id)
			throws JsonGenerationException, JsonMappingException, IOException {
		ObjectMapper mapper = MetaDataRepository.getSearchResultObjectMapper();
		StringWriter sw = new StringWriter();
		mapper.writeValue(sw, content);
		String contentString = sw.toString();
		addContentToIndex(contentString, id.toString());
	}

	private static final Charset UTF8 = Charset.forName("UTF-8");

	private void addContentToIndex(String content, String id) {
		RestTemplate restTemplate = new RestTemplate();
		HttpHeaders headers = new HttpHeaders();
		headers.setContentType(MediaType.APPLICATION_JSON);
		HttpEntity<String> request = new HttpEntity<String>(content, headers);
		String encodedId = URLEncoder.encode(id, UTF8);
		encodedId = id.replace("/", "%2F");
		URI uri;
		try {
			uri = new URI("http://localhost:29200/metadata/_doc/" + encodedId);
			restTemplate.put(uri, request);
		} catch (URISyntaxException e) {
			LOGGER.warn("Cannot add " + id, e);
		}

	}

	private ExerciseId calculateProjectId(String relativePath) {
		int relativePathPos = relativePath.indexOf("/");
		if (relativePathPos < 0) {
			int projectId = Integer.parseInt(relativePath);
			return new ExerciseId(projectId + "", null);
		} else {
			int projectId = Integer.parseInt(relativePath.substring(0, relativePathPos));
			return new ExerciseId(projectId + "", relativePath.substring(relativePathPos + 1));
		}
	}

	private UserProvidedMetadataDTO parseMetaDataFile(File metaDataFile)
			throws JsonParseException, JsonMappingException, IOException {
		ObjectMapper mapper = new ObjectMapper(new YAMLFactory());
		mapper.findAndRegisterModules();
		UserProvidedMetadataDTO result = mapper.readValue(metaDataFile, UserProvidedMetadataDTO.class);
		return result;
	}

	private ExtendedGitProject parseProjectFile(File projectFile)
			throws JsonParseException, JsonMappingException, IOException {
		ObjectMapper mapper = new ObjectMapper(new YAMLFactory());
		mapper.findAndRegisterModules();
		ExtendedGitProject result = mapper.readValue(projectFile, ExtendedGitProject.class);
		return result;
	}

	private interface Harvester<T> {
		List<T> harvest(File f, List<T> subResults, String relativePath, ExtendedGitProject gitProject);
	}

	/**
	 * recursively visits all nodes, collects all children metadta and then calls
	 * the harvester
	 * 
	 * @param <T>          the type the harvester returns
	 * @param dir          the directory to start harvesting in
	 * @param harvester    the harvester function
	 * @param relativePath the relative path to dir
	 * @param gitProject   the gitProject that may be collected from root.
	 * @return
	 */
	private <T> List<T> visitSubdirectories(File dir, Harvester<T> harvester, String relativePath,
			ExtendedGitProject gitProject) {
		Assert.state(dir.exists() && dir.isDirectory(), "must be a directory");
		List<T> harvestedMetadata = new ArrayList<T>();
		for (File containedFile : dir.listFiles()) {
			if (!containedFile.isDirectory() && containedFile.getName().equals("project.yaml")) {
				if (gitProject != null) {
					LOGGER.error("A project.yaml should not overwrite an existing project in a subnode! {}",
							containedFile.getAbsolutePath());
				}
				try {
					gitProject = parseProjectFile(containedFile);
				} catch (IOException e) {
					LOGGER.error("Cannot parse project.yaml for " + containedFile.getAbsolutePath(), e);
				}
			}
		}
		for (File containedFiles : dir.listFiles()) {
			if (containedFiles.isDirectory())
				harvestedMetadata.addAll(visitSubdirectories(containedFiles, harvester,
						(StringUtils.isEmpty(relativePath) ? "" : (relativePath + "/")) + containedFiles.getName(),
						gitProject));
		}
		for (File containedFiles : dir.listFiles()) {
			if (!containedFiles.isDirectory() && containedFiles.getName().equals("metaData.yaml")) {

				harvestedMetadata
						.addAll(harvester.harvest(containedFiles, harvestedMetadata, relativePath, gitProject));
				return harvestedMetadata;
			}
		}
		return harvestedMetadata;
	}

	/**
	 * replaces existing metadata by new content
	 * 
	 * @param content
	 * @param id
	 */
	public void updateMetadata(String content, int id) {
		RestTemplate restTemplate = new RestTemplate();
		HttpHeaders headers = new HttpHeaders();
		headers.setContentType(MediaType.APPLICATION_JSON);
		HttpEntity<String> request = new HttpEntity<String>(content, headers);
		restTemplate.put("http://localhost:29200/metadata/_doc/" + id, request, String.class);
	}

	public String updateMetadata(SearchResultDTO testExercise, int id)
			throws JsonGenerationException, JsonMappingException, IOException {

		// first fetch previous content
		RestTemplate restTemplate = new RestTemplate();
		String previousContent = restTemplate.getForObject("http://localhost:29200/metadata/_doc/" + id, String.class);

		ObjectMapper objectMapper = new ObjectMapper();
		final JavaTimeModule javaModule = new JavaTimeModule();

		objectMapper.registerModule(javaModule);
		objectMapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);
		StringWriter sw = new StringWriter();
		objectMapper.writeValue(sw, testExercise);
		final String contentString = sw.toString();
		updateMetadata(contentString, id);

		return previousContent;
	}

	/**
	 * sets the content to a raw json string (containing "_source" attribute). Used
	 * to write back a changed value from updateMetadata(...).
	 * 
	 * @param rawInput the json string
	 * @param id       the id to write to
	 * @throws JsonGenerationException
	 * @throws JsonMappingException
	 * @throws IOException
	 */
	public void resetRawMetadata(String rawInput, int id)
			throws JsonGenerationException, JsonMappingException, IOException {

		ObjectMapper objectMapper = new ObjectMapper();
		final JavaTimeModule javaModule = new JavaTimeModule();

		objectMapper.registerModule(javaModule);
		objectMapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);

		final JsonNode tree = objectMapper.readTree(rawInput);
		final JsonNode source = tree.at("/_source");
		StringWriter sw = new StringWriter();
		objectMapper.writeValue(sw, source);
		final String contentString = sw.toString();
		updateMetadata(contentString, id);

	}

	/**
	 * we have to wait until correct startup of content indexing?
	 * 
	 * @param minExpectedHits
	 */
	private void waitForCleanStartUp(int minExpectedHits) {
		RestHighLevelClient client = new RestHighLevelClient(RestClient.builder(new HttpHost("localhost", 29200)));
		int tryCount = 0;
		while (tryCount < 10) {
			tryCount++;
			SearchRequest searchRequest = new SearchRequest(SearchRepositoryConstants.INDEX_METADATA);
			BoolQueryBuilder queryBuilder = QueryBuilders.boolQuery();
			SearchSourceBuilder sourceBuilder = new SearchSourceBuilder();
			sourceBuilder.query(queryBuilder);
			SearchResponse searchResponse;
			try {
				searchResponse = client.search(searchRequest, RequestOptions.DEFAULT);
				if (searchResponse.getHits().getTotalHits() >= minExpectedHits) {
					LOGGER.info("Metadata index is up with {} hits after {} trials",
							searchResponse.getHits().getTotalHits(), tryCount);
					return;
				} else {
					LOGGER.warn("Metadata index only {} hits up after {} trials",
							searchResponse.getHits().getTotalHits(), tryCount);
				}
			} catch (IOException e) {
				LOGGER.warn("Metadata index startup exception", e);
			}
			try {
				Thread.sleep(1000L);
			} catch (InterruptedException e) {
				LOGGER.debug("Never mind: {}", e.getMessage());
			}
		}
	}

	private String getMetaDataConfigDefinition() throws IOException {
		InputStream metaDataStream = this.getClass().getResourceAsStream("../testData/es_metadata.schema.json");
		ByteSource byteSource = new ByteSource() {

			@Override
			public InputStream openStream() throws IOException {
				return metaDataStream;
			}
		};
		String text = byteSource.asCharSource(Charsets.UTF_8).read();
		return text;
	}

	@EventListener
	public void onApplicationEvent(ApplicationStartingEvent event) {
		try {
			startTestNode();
		} catch (IOException | NodeValidationException e) {
			LOGGER.error("Cannot startup TestNode", e);
		}
	}

	@EventListener
	public void onApplicationEvent(ContextStartedEvent event) {
		try {
			startTestNode();
		} catch (IOException | NodeValidationException e) {
			LOGGER.error("Cannot startup TestNode", e);
		}
	}

	@EventListener
	public void onApplicationEvent(ContextStoppedEvent event) {
		try {
			testNode.close();
		} catch (IOException e) {
			LOGGER.error("Cannot startup TestNode", e);
		}
	}

	public static class TestESNode extends Node {
		public TestESNode(Settings preparedSettings, Collection<Class<? extends Plugin>> classpathPlugins) {
			super(InternalSettingsPreparer.prepareEnvironment(preparedSettings, null), classpathPlugins, true);
		}

		@Override
		protected void registerDerivedNodeNameWithLogger(String nodeName) {
			LogManager.getLogger(nodeName);
		}
	}

}
